package net.sf.corn.converter.json;

import static net.sf.corn.converter.json.IJsonConstants.*;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Date;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import net.sf.corn.converter.ConversionException;
import net.sf.corn.converter.ObjectReferenceMap;

/**
 * This class is designed to be a generic JSON converter which covers almost all Java types.
 * Although this converter supports most java class types, it's quite easy
 * to register your own converter classes on runtime by using registerTransformer method of this class
 * to improve conversion performance. Another way of registering your own converters is putting them 
 * into "jsonConverters.properties" file in the root of the classpath. This file contains classic key,value
 * pairs in where key is a specifier and value is the classname of the converter. It's expected that 
 * classes implement {@link IJsonConverter} interface .
 *  
 * @author Serhat Dirik
 *
 */
public final class JsonConverter implements IJsonConverter {

	private static Map<Class<?>,Class<? extends IJsonConverter>> subjectMatters = new HashMap<>();
	private static Class<? extends IJsonConverter> beanConverter=null;
	private static Class<? extends IJsonConverter> objectConverter=null;
	private static Class<? extends IJsonConverter> enumConverter=null;
	private static Class<? extends IJsonConverter> arrayConverter=null;
	private static List<Class<?>> convertibles = new ArrayList<>();
	
	private final static String PRIMITIVE_BOOLEAN="boolean"; 
	private final static String PRIMITIVE_CHAR="char";
	private final static String PRIMITIVE_INT="int";
	private final static String PRIMITIVE_DOUBLE="double";
	private final static String PRIMITIVE_FLOAT="float";
	private final static String PRIMITIVE_BYTE="byte";
	private final static String PRIMITIVE_SHORT="short";
	private final static String PRIMITIVE_LONG="long";
	private final static String JSON_CONVERTERS_PROPS = "jsonConverters.properties";
	
	static {
		registerConverter(CnvChar.class);
		registerConverter(CnvString.class);
		registerConverter(CnvDate.class);
		registerConverter(CnvCalendar.class);
		registerConverter(CnvNumeric.class);
		registerConverter(CnvList.class);
		registerConverter(CnvSet.class);
		registerConverter(CnvMap.class);
		registerConverter(CnvPattern.class);
		arrayConverter = CnvArray.class;
		enumConverter = CnvEnum.class;
		beanConverter = CnvBean.class;
		objectConverter = CnvObject.class;
		registerExternalConverters();
	}
	/**
	 * Registers 3th party converters from the jsonConverters.properties in the classpath root
	 */
	private static void registerExternalConverters(){
		ClassLoader loader = Thread.currentThread().getContextClassLoader();
		InputStream inputStream = loader.getResourceAsStream(JSON_CONVERTERS_PROPS);
		if(inputStream == null) return;
		Properties props = new Properties();
		try {
			props.load(inputStream);
			for(String key: props.stringPropertyNames()){
				String className = props.getProperty(key);
				try {
					@SuppressWarnings("unchecked")
					Class<? extends IJsonConverter> clazz = (Class<? extends IJsonConverter>) Class.forName(className);
					registerConverter((Class<? extends IJsonConverter>) clazz);
				} catch (ClassNotFoundException | ClassCastException e) {
				}
			}
		} catch (IOException e) {
		}
	}
	
	private static List<Class<?>> getConvertibles() {
		return convertibles;
	}

	/**
	 * Registers a converter to be used for the specified types within it
	 * @param clazz Converter class that implements IJsonConverter
	 */
	public static final synchronized  void registerConverter(Class<? extends IJsonConverter> clazz)  {
		if (clazz == null){
			throw new IllegalArgumentException("Null transformer can't be registered!");
		}
		try {
			IJsonConverter converter = clazz.newInstance();
			for (Class<?> subject : converter.findConvertibles()) {
				getConvertibles().add(subject);
				subjectMatters.put(subject,clazz);
			} 
		} catch (InstantiationException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}
	/**
	 * Registers a converter to be used with specified subject only
	 * @param clazz Converter class that implements IJsonConverter
	 * @param subject Subject class 
	 */
	public static final synchronized  void registerConverterForSubject(Class<? extends IJsonConverter> clazz,Class<?> subject)  {
		if (clazz == null){
			throw new IllegalArgumentException("Null transformer can't be registered!");
		}
		if (subject == null){
			throw new IllegalArgumentException("Null subject can't be registered!");
		}
		getConvertibles().add(subject);
		subjectMatters.put(subject,clazz);
	}
	


	
	private IJsonConverter findRegisteredConverter(Class<?> subject) throws ConversionException {
		if(subject.isAnnotation()){
			throw new ConversionException("Conversion not supported for annotations !"+subject.getName());
		}
		if(subject.isArray()){
			try {
				return arrayConverter.newInstance();
			} catch (InstantiationException e) {
				throw new ConversionException(e);
			} catch (IllegalAccessException e) {
				throw new ConversionException(e);
			}
		}
		Class<? extends IJsonConverter> converterClazz = subjectMatters.get(subject);
		if(converterClazz == null){
			if(converterClazz == null && subject.isEnum()){
				converterClazz =  enumConverter;	
				IJsonConverter cnvObj;
				try {
					cnvObj = converterClazz.newInstance();
				} catch (InstantiationException e) {
					throw new ConversionException(e);
				} catch (IllegalAccessException e) {
					throw new ConversionException(e);
				}
					return cnvObj;
			}else if (converterClazz == null){
				for(Class<? extends IJsonConverter> clazz:subjectMatters.values()){
					IJsonConverter cnvObj;
					try {
						cnvObj = clazz.newInstance();
					} catch (InstantiationException e) {
						throw new ConversionException(e);
					} catch (IllegalAccessException e) {
						throw new ConversionException(e);
					}
					if (cnvObj.isConvertible(subject)){
						converterClazz = clazz;
						registerConverterForSubject(clazz, subject);
						return cnvObj;
					}
				}
			}
			converterClazz = beanConverter;
		} 
		
		try {
			IJsonConverter converter = converterClazz.newInstance();
			if (! converter.isConvertible(subject)){
				converter = objectConverter.newInstance();
				if (! converter.isConvertible(subject)){
				 throw new ConversionException("Transformation not supported for class "+subject.getName());
				}
			}
			return converter;
		} catch (InstantiationException e) {
			throw new ConversionException(e);
		} catch (IllegalAccessException e) {
			throw new ConversionException(e);
		}
		 
	}
	@Override
	public   Object toJava(JsTypeObject obj)
		throws ConversionException {
		return toJava(null,obj);
	}
	
	@Override
	public  Object toJava(Class<?> clazz, JsTypeObject obj)
			throws ConversionException {
		if (obj == null ||obj.isNull()) {
			return null;
		}
		
		
		if (clazz == null) {
			try{
				clazz = obj.getJavaClass();
			}catch(ClassNotFoundException e){
				throw new ConversionException(obj.getJavaClassName()+" not found in the classpath!",e);
			}catch(Exception e){
				throw new ConversionException(this.getClass().getName()
						+ " can't be used as converter for " + obj.getJavaClassName(),e);
			}
			if (clazz == null) {
				//Fallback to default classes
				if(obj.isSimple()){
					clazz = detectClass(obj.toStringValue());
				}else if(obj.isComplex()){
					JsTypeComplex complex = (JsTypeComplex) obj;
					if(complex.has(ATTR_TIME) && complex.size() <= 2){
						if(complex.has(ATTR_TIMEZONEID)){
							clazz = Calendar.class;
						}else if(complex.size() == 1){
							clazz = Date.class;
						}
					}
					if(clazz == null) clazz = Map.class;
				}else if(obj.isList()){
					clazz = List.class;
				}
			}
		}
		
		IJsonConverter converter = findRegisteredConverter(clazz);
		if (converter == null) {
			throw new ConversionException(this.getClass().getName()
					+ " can't be used as converter for " + clazz.getName());
		}
		return converter.toJava(clazz,obj);
	}
	
	private static Class<?> detectClass(String str){
		try{
			Integer.valueOf(str.trim());
			return Integer.class;
		}catch(Exception ex){
		}
		try{
			Float.valueOf(str.trim());
			return Float.class;
		}catch(Exception ex){
		}
		try{
			Double.valueOf(str.trim());
			return Double.class;
		}catch(Exception ex){
		}
		try{
			new BigInteger(str.trim());
			return BigInteger.class;
		}catch(Exception ex){
		}
		try{
			new BigDecimal(str.trim());
			return BigDecimal.class;
		}catch(Exception ex){
		}
		return String.class;
	}
	
	@Override
	public  JsTypeObject toTargetType(Object o) throws ConversionException {
		return toTargetType(o, null);
	}

	@Override
	public  JsTypeObject toTargetType(Object o, ObjectReferenceMap referenceMap)
			throws ConversionException {
		if (o == null || 
			(o instanceof String && o.equals(JsTypeObject.NULL_STRING_VALUE) )) {
			
			return JsTypeNull.NULL;
		}
		if(o instanceof JsTypeObject){
			return (JsTypeObject) o;
		}
		IJsonConverter converter = findRegisteredConverter(o.getClass());
		if (converter == null) {
			throw new ConversionException(this.getClass().getName()
					+ " can't be used as converter for " + o.getClass().getName());
		}
		return converter.toTargetType(o, referenceMap);
	}


	@Override
	public List<Class<?>> findConvertibles() {
		return Collections.unmodifiableList(convertibles);
	}

	protected boolean checkIfConvertibleFromRegisteredTypes(Class<?> clazz) {
		for(Class<?> type:convertibles){
			if ((clazz == null && ! type.isPrimitive()) || (clazz != null && type.isAssignableFrom(clazz)) ) return true;
		}
		return false;
	}
	 
	public  boolean isConvertible(Class<?> clazz, JsTypeObject obj) {
		if (obj == null || obj.isNull()) {
			return true;
		}
		boolean ret = false;
		try {
			if(obj.getJavaClassName() != null){
				Class<?>  sourceClass = obj.getJavaClass();
				if(clazz.isAssignableFrom(sourceClass)){
					IJsonConverter converter;
					converter = findRegisteredConverter(clazz);
					if (converter == null) {
						ret = false;
					}else{
						ret = converter.isConvertible(sourceClass);
					}
				}
			}else{
				IJsonConverter converter;
				converter = findRegisteredConverter(clazz);
				if (converter == null) {
					ret = false;
				}else{
					//Throws an error if not converted
					 Object javaObj = converter.toJava(clazz,obj);
					 ret = javaObj != null;			
				}
			}
			
		} catch (Exception e) {
			ret= false;
		}
		return ret;
	}
	
	@Override
	public boolean isConvertible(String className) {
		if (className == null || className.trim().length() == 0) {
			return true;
		}
		Class<?> clazz;
		if(className.trim().equals(PRIMITIVE_BOOLEAN)){
			clazz = boolean.class;
		}else if(className.trim().equals(PRIMITIVE_BYTE)){
			clazz = byte.class;
		}else if(className.trim().equals(PRIMITIVE_CHAR)){
			clazz = char.class;
		}else if(className.trim().equals(PRIMITIVE_DOUBLE)){
			clazz = double.class;
		}else if(className.trim().equals(PRIMITIVE_FLOAT)){
			clazz = float.class;
		}else if(className.trim().equals(PRIMITIVE_INT)){
			clazz = int.class;
		}else if(className.trim().equals(PRIMITIVE_LONG)){
			clazz = long.class;
		}else if(className.trim().equals(PRIMITIVE_SHORT)){
			clazz = short.class;
		}else {
			try {
				
				clazz = Class.forName(className);
			} catch (ClassNotFoundException e) {
				//e.printStackTrace();
				return false;
			}
		}
		
		return isConvertible(clazz);
	}

	@Override
	public boolean isConvertible(Class<?> clazz) {
		return checkIfConvertibleFromRegisteredTypes(clazz);
	}

}
